!=========================================================================
! Copyright (C) VMware, Inc. 1986-2011.  All Rights Reserved.
!
! $Id: classorganizer.gs 26764 2011-09-08 21:02:26Z otisa $
!
! Superclass Hierarchy:
!   ClassOrganizer, Object.
!
!=========================================================================

expectvalue %String
run

  Object _newKernelSubclass: 'ClassOrganizer'
    instVarNames: #(#classes #classNames #user #hierarchy #categories
                    #rootClass)
    classVars: #()
    classInstVars: #()
    poolDictionaries: { }
    inDictionary: Globals
      options: #()
    reservedOop: 729
%

removeallmethods ClassOrganizer
removeallclassmethods ClassOrganizer

category: 'For Documentation Installation only'
classmethod: ClassOrganizer
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'A ClassOrganizer collects classes from the current user''s symbol list
 and organizes them into searchable tables that allow tools to present
 the classes and to perform cross-referencing and fileout.

 An organizer can also be created to work with a subtree of another
 organizer''s hierarchy.  Such organizers do not track categorization
 of classes but only the subtree of the overall hierarchy.'.
doc documentClassWith: txt.

txt := (GsDocText new) details:
'A ClassSet of all the classes found by an instance.'.
doc documentInstVar: #classes with: txt.

txt := (GsDocText new) details:
'Class name information.'.
doc documentInstVar: #classNames with: txt.

txt := (GsDocText new) details:
'Reserved for future use.'.
doc documentInstVar: #user with: txt.

txt := (GsDocText new) details:
'An IdentityDictionary of class->subclasses associations.'.
doc documentInstVar: #hierarchy with: txt.

txt := (GsDocText new) details:
'A Dictionary of category->classes associations.'.
doc documentInstVar: #categories with: txt.

txt := (GsDocText new) details:
'The root class of the instance.'.
doc documentInstVar: #rootClass with: txt.

self description: doc.
%

! ------------------- Class methods for ClassOrganizer
category: 'Instance Creation'
classmethod: ClassOrganizer
new

"Creates and returns a new instance of ClassOrganizer with a root of Object."

| inst |

inst := (super new) rootClass: Object.
inst updateClassInfo.
^inst
%

category: 'Instance Creation'
classmethod: ClassOrganizer
newWithRoot: aClass from: anotherOrganizer

"Creates a new ClassOrganizer that is limited to the given subtree of objects."

| inst |
inst := super new.
inst rootClass: aClass.
inst classes: ClassSet new.
inst classes add: aClass.
inst classes addAll: (anotherOrganizer allSubclassesOf: aClass).
inst rebuildHierarchy.
^inst
%

category: 'Instance Creation'
classmethod: ClassOrganizer
newWithRoot: aClass forUserId: aUserId

"Creates a new ClassOrganizer that is limited to the given subtree of objects using
 the SymbolList from aUserId.  Caller must have read access to given user's SymbolList."

| inst u |
u := AllUsers userWithId: aUserId .
inst := super new .
inst rootClass: aClass.
inst classes: ClassSet new.
inst classes add: aClass.
inst collectClassesFromSymbolList: u symbolList;
     rebuildHierarchy;
     collectClassNames;
     rebuildCategories .
^inst
%

category: 'Updating'
classmethod: ClassOrganizer
clearCachedOrganizer
  "do nothing in base image"
  ^ self
%

! ------------------- Instance methods for ClassOrganizer
category: 'Private - Class Collection'
method: ClassOrganizer
addCachedClassNames

"Adds new class names to the auto-complete set.  New class names are cached
 until the auto-completer is needed, then they are all merged in to the
 completer in one shot.  Users that don't make use of auto-completion features
 never have to pay the price of updating the completer's structures."

| completer cache |
completer := classNames at: 1.
cache := classNames at: 2.
1 to: cache size do: [:i |
  completer addString: (cache at: i).
].
classNames at: 2 put: nil
%

category: 'Updating'
method: ClassOrganizer
addClass: cls

"Adds the class cls, replacing any existing class with the same superclass."

| nm superClass cat old matching |

hierarchy == nil ifTrue: [^self].
nm := cls name.
superClass := cls superClass.
superClass == nil ifTrue:[ ^ self "subclasses of nil not supported, bug13156"].

(hierarchy includesKey: superClass) ifTrue: [
  hierarchy at: superClass put:
    ((hierarchy at: superClass) select: [:e |
        e name ~= nm]).
  (hierarchy at: superClass) add: cls.
]
ifFalse: [ | cset |
  (cset := ClassSet new) add: cls .
  hierarchy at: superClass put: cset .
].

matching := classes select: {:each | each.name = nm}.
matching size > 0 ifTrue: [
  classes := classes - matching
].
classes add: cls.

cat := categories at: cls category otherwise: nil.
cat == nil ifTrue: [
  categories at: cls category put: { cls }.
]
ifFalse: [
  nm := cls name.
  [old := cat findFirst: [:x | x name = nm].
    old > 0] whileTrue: [
    cat removeFrom: old to: old
  ].
  cat add: cls
].

self addedClassName: nm
%

category: 'Private - Class Collection'
method: ClassOrganizer
addedClassName: name

"Adds the new class name to the auto-complete set."

(classNames at: 2) == nil ifTrue: [
  classNames at: 2 put: { } .
].
(classNames at: 2) add: name.
%

category: 'Reporting'
method: ClassOrganizer
allReferencesTo: selector

"Returns an Array of two Arrays.  The first contains GsMethods that
 implement, send, or refer to the given selector.  The second contains the
 indexes into sourceStrings where the first reference takes place."

^self allReferencesTo: selector in: classes
%

category: 'Reporting'
method: ClassOrganizer
allReferencesTo: aSelector in: classSet

"Returns an Array of two Arrays.  The first contains GsMethods that
 implement, send, or refer to the given selector.  The second contains the
 indexes into sourceStrings where the first reference takes place."

| methods indices cset selectorSym methsSet |

methsSet := IdentitySet new .
methods := { } .
indices := { } .
selectorSym := Symbol _existingWithAll: aSelector .
selectorSym ifNotNil:[
  "Sort the class set by name, then search"
  cset := classSet sortAscending.
  1 to: cset size do: [ :i |
    1 to: 2 do: [ :n | | cls mDict |
      cls := (n == 1 ifTrue: [cset at: i] ifFalse: [cls class]).
      mDict := cls _fullMethodDictEnv0 .
      mDict valuesDo: [ :method | | srcOffset |
	  "Check for implementors first"
	  method selector == selectorSym ifTrue: [
	    (methsSet _addIfAbsent: method) ifTrue:[
	      methods add: method.
	      indices add: 1.
	    ].
	  ] ifFalse: [
	    "Check for senders"
	    srcOffset := method _sourceOffsetOfFirstSendOf: selectorSym .
	    srcOffset ifNotNil:[
	      (methsSet _addIfAbsent: method) ifTrue:[
		methods add: method .
		indices add: srcOffset.
	      ].
	    ] ifNil: [ 
	      (method _literalsIncludesSymbol: selectorSym value: nil) ifTrue:[
		(methsSet _addIfAbsent: method) ifTrue:[
		  methods add: method.
		  indices add: (method sourceString findString: selectorSym startingAt: 1).
		]. 
	      ].
	    ].
	  ].
      ].
    ].
  ].
].
^{ methods . indices }
%

category: 'Queries'
method: ClassOrganizer
allSubclassesOf: aClass

"Returns a collection of all the subclasses of the given class: an Array that
 holds a depth-first traversal of the class hierarchy subtree rooted at
 aClass."

| result subs |

classes == nil ifTrue: [
  self updateClassInfo
].
(classes includesIdentical: aClass) ifFalse: [
  self addClass: aClass.
  ^self allSubclassesOf: aClass
].

result := { } .
subs := hierarchy at: aClass otherwise: nil.
subs size > 0 ifTrue: [
  subs := subs sortAscending.
  subs do: [:each |
    result add: each; addAll: (self allSubclassesOf: each)
  ].
].
^result
%

category: 'Accessing'
method: ClassOrganizer
categories

"Returns the value of the instance variable 'categories', which are the
 class categories."

^categories
%

category: 'Queries'
method: ClassOrganizer
categoryCrossReference

"Returns a dictionary of all method categories and the classes with methods
 in each category."

| resultDict cls |

resultDict := SymbolDictionary new.
1 to: classes size do: [ :i |
  1 to: 2 do: [ :which |
    cls := classes _at: i.
    which == 2 ifTrue: [ cls := cls class ].
    cls categorysDo:[ :aCateg :selectorSet | | clsset |
        clsset := resultDict at: aCateg otherwise: nil.
        clsset == nil ifTrue: [
	  clsset := { }.
	  resultDict at: aCateg put: clsset.
        ].
        clsset add: cls.
    ].
  ].
].
^resultDict
%

category: 'Reports'
method: ClassOrganizer
categoryCrossReferenceByName

"Returns a String containing a report from a cross-reference of method
 categories. For each method category, the report contains a list of
 names of classes which have methods in that category."

^self categoryCrossReferenceReportByName: nil
%

method: ClassOrganizer
methodCategories
  "Returns a String containing a report of all the method categories"
| report cats |
cats := self _methodCategories sortAscending:'' .
report := String new.
cats do:[ :aCateg| report add: aCateg ; lf ].
^ report
%

method: ClassOrganizer
methodsInCategory: aString
  "Return a String containing a report of all methods in the specified category"

^ self _asReportString:( self _methodsInCategory: aString) 
%

method: ClassOrganizer
methsInCat: aString
  "Return a String containing a report of all methods in the specified category"
^ self methodsInCategory: aString
%

category: 'Private'
method: ClassOrganizer
_methodsInCategory: aString 
| catSym sortedClasses methsSet methsArr |
catSym := aString asSymbol .
methsSet := IdentitySet new .
methsArr := { } .
sortedClasses := classes sortAscending .
1 to: sortedClasses size do: [ :i |
  1 to: 2 do: [ :which | | cls sels |
    cls := sortedClasses _at: i.
    which == 2 ifTrue: [ cls := cls class ].
    sels := cls selectorsIn: catSym .
    sels size ~~ 0 ifTrue:[ | mDict |
      mDict := cls _fullMethodDictEnv0 .
      sels do:[ :aSel | | meth |
          (meth := mDict at: aSel otherwise: nil) ifNotNil:[
             (methsSet _addIfAbsent: meth) ifTrue:[ methsArr add: meth ]
          ].
      ].
    ].
  ].
].
^ methsArr
%
     
method: ClassOrganizer
_methodCategories
 "Returns an IdentitySet of all method categories"
| set |
set := IdentitySet new .
1 to: classes size do: [ :i |
  1 to: 2 do: [ :which | | cls |
    cls := classes _at: i.
    which == 2 ifTrue: [ cls := cls class ].
    cls categorysDo:[ :aCateg :selectors |  set add: aCateg ].
  ].
].
^ set
%

method: ClassOrganizer
categoryCrossReferenceReportByName: catsDict

"Returns a String containing a report from a cross-reference of categories.
 For each method category, the report contains a list of
 names of classes which have methods in that category.

 The argument may be nil or a dictionary similar to the one that
 categoryCrossReference returns."

  | dict report cats |
(dict := catsDict) == nil ifTrue: [
  dict := self categoryCrossReference
].
report := String new.
cats := SortedCollection withAll: dict keys.
1 to: cats size do: [ :i | | cat clsset |
  cat := cats at: i.
  report add: cat; add: Character tab.
  clsset := dict at: cat.
  1 to: clsset size do: [ :j |
    report add: (clsset at: j) name.
    j < clsset size ifTrue: [ report add: $, ].
  ].
  report lf .
].
^report .
%

category: 'Accessing'
method: ClassOrganizer
classCompletion

"Returns the AutoComplete holding the class names."

(classNames at: 2) ~~ nil ifTrue: [
  self addCachedClassNames
].
^(classNames at: 1).
%

category: 'Accessing'
method: ClassOrganizer
classNames

"Returns the Array of class names held by the receiver."

(classNames at: 2) ~~ nil ifTrue: [
  self addCachedClassNames
].
^(classNames at: 1) strings.
%

category: 'Accessing'
method: ClassOrganizer
classes

"Returns the ClassSet of classes held by the receiver."

^classes
%

category: 'Updating'
method: ClassOrganizer
classes: aClassSet

"Updates the set of classes held by the receiver.  The receiver's
 hierarchy should be rebuilt after this (see rebuildHierarchy)."

classes := aClassSet
%

category: 'Private - Class Collection'
method: ClassOrganizer
collectClassNames

"Causes the receiver to collect all of the names of classes from the
 current collection of classes and form an auto-completer for the names."

| arr sz |

sz := classes size .
arr := Array new: sz .
1 to: classes size do: [:i | arr at: i put:((classes _at: i) name)  ].

classNames := { AutoComplete new strings: arr cluster: false . nil }.
%

category: 'Private'
method: ClassOrganizer
symbolList

"Returns the symbol list of the current user."

^ System myUserProfile symbolList copy
%

category: 'Private'
method: ClassOrganizer
dictionaryAndSymbolOf: anObject

^ System myUserProfile dictionaryAndSymbolOf: anObject
%

category: 'Private - Class Collection'
method: ClassOrganizer
collectClassesFromSymbolList: aSymbolList

"Rebuilds the class hierarchy structure from the given SymbolList."

<primitive: 2001> "enter protected mode to be able to modify symbol list"
| prot |
prot := System _protectedMode .
[
  | allClasses symlist dict rootIsObj done |
  allClasses := ClassSet new.

  " make a list of all the named classes "

  symlist := aSymbolList copy.
  rootIsObj := rootClass == Object.
  done := IdentitySet new.

  [symlist size > 0] whileTrue: [
    dict := symlist at: 1.
    (done includesIdentical: dict) ifFalse: [
      dict valuesDo: [:aValue | | anObj |
	anObj:= aValue.
	anObj isBehavior ifTrue: [
	  anObj isMeta ifTrue: [ anObj := anObj thisClass ].
	  (rootIsObj or: [anObj isSubclassOf: rootClass]) ifTrue: [
	    allClasses add: anObj
	  ].
	].
      ].
      done add: dict.
    ].
    symlist protectedRemoveFrom: 1 to: 1.
  ].
  classes := allClasses.
] ensure:[
  prot _leaveProtectedMode
]
%

category: 'Private - Class Collection'
method: ClassOrganizer
collectClasses

"Rebuilds the class hierarchy structure.  This message should be sent whenever
 new classes have been created or imported from other users."

^ self collectClassesFromSymbolList: self symbolList
%

category: 'Fileout Aids'
method: ClassOrganizer
determineClassFileoutOrder: classdict

"Returns an ordered collection of the values that are classes in classdict,
 specifying the order of fileout.  The argument should be a SymbolDictionary."

| org order classSet block |

( classdict isEmpty or:
[ nil == (classdict detectAssociations: [ :assoc | assoc value isBehavior ] ifNone: [ nil ]) ])
  ifTrue: [ ^ { } ].

"form a hierarchy for the set of classes"
org := Dictionary new.
org at: #nil put: ClassSet new.

classSet := ClassSet new.
classdict associationsDo: [:assn | | sub |
  sub := assn value.
  sub isBehavior ifTrue: [ | superCls |
    classSet add: sub.
    [superCls := sub superClass.  superCls ~~ nil] whileTrue: [ | assoc |
      assoc := org associationAt: superCls otherwise: nil.
      assoc == nil ifTrue: [
        assoc := Association newWithKey: superCls value: ClassSet new.
          org add: assoc
      ].
      assoc value add: sub.
      sub := superCls
    ].
    (org at: #nil) add: sub.
  ].
].

"make a recursive block to order the subclass sets and weed out unwanted 
 classes"
order := { } .
block := [:order:subs |
  1 to: subs size do: [:i | | assoc class |
    class := subs at: i.
    (classSet includesIdentical: class) ifTrue: [
      order add: class
    ].
    assoc := org associationAt: class otherwise: nil.
    (assoc ~~ nil and: [assoc value size > 0]) ifTrue: [
      block value: order value: (assoc value sortAscending)
    ]
  ]
].

block value: order value: ((org at: #nil) sortAscending).
^order
%

category: 'Fileout Aids'
method: ClassOrganizer
fileOutClassesAndMethodsInDictionary: aSymbolDictionary on: aStream

"Files out all source code for classes in aSymbolDictionary in Topaz filein
 format on aStream."

| order |
order := self determineClassFileoutOrder: aSymbolDictionary.
self fileOutClasses: order on: aStream inDictionary: aSymbolDictionary
     named: (aSymbolDictionary name).
self fileOutMethods: aSymbolDictionary order: order on: aStream
%

category: 'Fileout Aids'
method: ClassOrganizer
fileOutClasses: order on: stream inDictionary: dict named: dictName

"Writes out code on the given stream that creates the given classes in the
 dictionary with the given name.  The dict argument should be a SymbolDictionary
 of classes."

| class str lf head term any nm |

order size == 0 ifTrue: [ ^self ].

head := 'doit
'.
term := '
' , '%
'.

lf := Character lf.

"Determine the dictionary name to use"
nm := dictName.

1 to: order size do: [:j |
  class := order at: j.
  class fileOutPreClassOn: stream.
  stream nextPutAll: head;
          nextPutAll:
    (class _modifiableDefinitionInDictionary: dict named: nm );
    nextPut: $.;
    nextPutAll: term
].

"now write out constraints and make non-modifiable classdict non-modifiable "
1 to: order size do: [:k |
  class := order at: k.
  any := false.
  str := class _constraintCreationExpressionIn: dict.
  str size > 0 ifTrue: [
    stream nextPutAll: head.
    stream nextPutAll: (dict keyAtValue: class); nextPut: $ ;
      nextPutAll: str;
      nextPut: $.; nextPut: lf.
    any := true.
  ].
  class isModifiable ifFalse: [
    any ifFalse: [
      stream nextPutAll: head.
      any := true
    ].
    stream nextPutAll: (dict keyAtValue: class); nextPutAll: ' immediateInvariant.'.
  ].
  any ifTrue: [
    stream nextPutAll: term
  ]
].
%

category: 'Private'
method: ClassOrganizer
addTempDictionary: classdict

System myUserProfile symbolList insertObject: classdict at: 1
%

category: 'Private'
method: ClassOrganizer
removeTempDictionary: classdict

System myUserProfile symbolList remove: classdict
%

category: 'Fileout Aids'
method: ClassOrganizer
fileOutMethods: classdict order: order on: stream

"File out each class's code and embedded classes."

| class |
classdict size == 0 ifTrue: [ ^self ].
"put the class dictionary in the temporary symbol list so it will be used for
  name resolution during method fileout"
self addTempDictionary: classdict.
1 to: order size do: [:l |
  class := order at: l.
  class fileOutCategoriesOn: stream.
].
self removeTempDictionary: classdict.
%

category: 'Fileout Aids'
method: ClassOrganizer
fileOutOtherMethods: methodInfo on: stream

"Files out a set of methods on the given stream/file.  methodInfo must be an
 Array of pairs: #(class selector)."

| pair |
1 to: methodInfo size do: [:i |
  pair := methodInfo at: i.
  "use the fileOutMethod: version here so the category will be included"
  stream nextPutAll: ((pair at: 1) fileOutMethod: (pair at: 2)).
].
%

category: 'Accessing'
method: ClassOrganizer
hierarchy

"Returns the value of the instance variable 'hierarchy'."

^hierarchy
%

category: 'Reporting'
method: ClassOrganizer
implementorsOf: aSelector

"Returns a collection of GsMethods that implement the given selector."

^self implementorsOf: aSelector in: classes
%

category: 'Reporting'
method: ClassOrganizer
implementorsOf: aSelector in: aclassSet

"Returns a collection of GsMethods that implement the given selector."

| result cset methsSet |

methsSet := IdentitySet new .
result := { } .
cset := aclassSet sortAscending.
1 to: cset size do: [ :j | | cls mDict meth |
  2 timesRepeat:[
    cls == nil ifTrue:[ cls := cset at: j ] ifFalse:[ cls := cls class].
    mDict := cls _fullMethodDictEnv0 .
    meth := mDict at: aSelector otherwise: nil .
    meth ifNotNil: [ (methsSet _addIfAbsent: meth) ifTrue:[ result add: meth ]].
  ]. 
].
^result
%

category: 'Reporting'
method: ClassOrganizer
implementorsOfReport: aSelector

"Returns a String describing the methods that are senders of the specified
 selector."

^ self _asReportString: (self implementorsOf: aSelector) indent: '' 
%

category: 'Reporting'
method: ClassOrganizer
impls: aSelector
  "a variant of implementorsOfReport: easier to type when using topaz ."
  ^ self implementorsOfReport: aSelector
%
method: ClassOrganizer
implsC: aSelector
  "a variant of implementorsOfReport: sorted by method categories"
  ^ self _asCategoriesReportString: (self implementorsOf: aSelector) 
%

category: 'Reporting'
method: ClassOrganizer
_asReportString: anArray indent: indentString
"Returns a String, one line per method in anArray."
| result LF arr |
arr := { } .
LF := Character lf .
1 to: anArray size do:[:j | | aMeth str |
  aMeth := anArray at: j .
  str := String withAll: aMeth inClass name .
  str addAll: ' >> ' ; addAll: aMeth selector .
  arr add: str .
].
arr := arr sortAscending:'' .
result := String new .
1 to: arr size do:[:j | 
  result add: indentString; add:(arr at:j) ; add: LF 
].
^ result
%
method: ClassOrganizer
_asCategoriesReportString: anArray 
"Returns a String. The result is sorted by method categories"
| lf dict cats rpt |
lf := Character lf .
dict := Dictionary new .
1 to: anArray size do:[:j | | aMeth str cls sel cat |
  aMeth := anArray at: j .
  cls := aMeth inClass .
  sel := aMeth selector .
  str := String withAll: cls name .
  str addAll: ' >> ' ; addAll: sel .
  cat := cls categoryOfSelector: sel environmentId: aMeth environmentId .
  ( dict at: cat ifAbsent:[ dict at: cat put: { } ]) add: str .
].
cats := dict sortAscending:'key' .
rpt := String new .
cats do:[:assoc | | arr |
  rpt add: assoc key ; add: lf .
  arr := assoc value sortAscending:'' .
  1 to: arr size do:[:j | 
    rpt add: '   '; add:(arr at:j) ; add: lf 
  ].
].
^ rpt
%

method:
_asReportString: anArray
  ^ self _asReportString: anArray indent: '' 
%

category: 'Private - Class Collection'
method: ClassOrganizer
rebuildCategories

""

| cls arr cat |

   "gemstone64,  changed categories from a SymbolDictionary 
    to a StringKeyValueDictionary to avoid unnecessary symbol creation"

categories := StringKeyValueDictionary new.
1 to: classes size do: [:i |
  cls := classes _at: i.
  cat := cls category.
  arr := categories at: cat otherwise: nil.
  arr == nil ifTrue: [
    arr := { }.
    categories at: cat put: arr
  ].
  arr add: cls
].
%

category: 'Private - Class Collection'
method: ClassOrganizer
rebuildHierarchy

"Builds an IdentityDictionary containing all classes as keys and ClassSets of
 their subclasses as values.  As a side-effect, the classes collection is
 expanded to include any superclasses that aren't in the user's name space."

| hier supers each sup supsup c |

hier := IdentityDictionary new.
supers := ClassSet new.

"add each class to its superclass's subclass set"
1 to: classes size do: [:i |
  each := classes _at: i.
  sup := each superClass.
  sup ~~ nil ifTrue: [
    (classes includesIdentical: sup) ifFalse: [
      "keep track of superclasses that aren't in the class set and complete
        their hierarchies on a second pass"
      supsup := sup.
      [ supsup ~~ nil ] whileTrue: [
        (supsup ~~ sup) ifFalse: [
          supers add: supsup
        ].
        supsup == rootClass ifTrue: [ supsup := nil ] ifFalse: [ supsup := supsup superClass ]
      ]
    ].
    (c := hier at: sup otherwise: nil) == nil ifTrue: [
      c := ClassSet new.
      hier at: sup put: c.
    ].
    c add: each
  ].
].

"get rid of superclasses that have already been processed"
supers := supers - classes.

"complete the class hierarchies for superclasses that weren't in the
  initial set of classes"
1 to: supers size do: [:i |
  each := supers _at: i.
  sup := each superClass.
  sup ~~ nil ifTrue: [
    (c := hier at: sup otherwise: nil) == nil ifTrue: [
      c := ClassSet new.
      hier at: sup put: c.
    ].
    c add: each
  ].
].

classes := classes + supers.
hierarchy := hier
%

category: 'Updating'
method: ClassOrganizer
recategorize: class to: newCategory

"Move the class from its present category to the given category."

| oldcat oldlist newlist idx |
oldcat := class category.
class category: newCategory.
oldlist := categories at: oldcat otherwise: nil.
oldlist ~~ nil ifTrue: [
  idx := oldlist indexOf: class.
  idx > 0 ifTrue: [
    oldlist removeFrom: idx to: idx
  ]
].
newlist := categories at: newCategory otherwise: nil.
newlist == nil ifTrue: [
  categories at: newCategory put: { class }
]
ifFalse: [
  newlist add: class
]
%

! _recompileWithSubclasses:newClass:  removed per bug 8052 .
!  this method sends messages that are not yet implemented in the server image
! 
! category: 'Schema Modification'
! method: ClassOrganizer
! _recompileWithSubclasses: oldClass newClass: newClass
! 
! "Copies information and methods from oldClass to newClass, replaces
!  oldClass with newClass in the organizer's structures, then does the same for
!  all of the subclasses of oldClass.  Returns an Array containing two
!  Arrays and a Boolean:
! 
!  1) the old classes
!  2) the corresponding new classes
!  3) whether or not all compilations were successful
! 
!  A browser on the methods that failed to recompile is opened to allow the user
!  to fix the problems.  The organizer is updated to contain the newly created
!  classes."
! 
! | cls oclass failedMethods failedStream failedText assn
!   oclasses nclasses badclasses sidx x pos dict autoLock badLocks lblock |
! 
! badclasses := Array new.
! 
! failedStream := WriteStream on: Array new.
! 
! failedMethods := newClass _copyMethodsAndVariablesFrom: oldClass
!   except: nil dictionaries: self symbolList.
! self linkOldClass: oldClass toNewClass: newClass.
! failedMethods size > 0 ifTrue: [
!   badclasses add: oldClass.
!   failedStream nextPutAll: failedMethods.
! ].
! 
! (classes includesIdentical: oldClass) ifFalse: [ oclasses := { } ]
! ifTrue: [
!   oclasses := self allSubclassesOf: oldClass.
! ].
! oclasses insertObject: oldClass at: 1.
! nclasses := Array with: newClass.
! oclasses size > 1 ifTrue: [
!   2 to: oclasses size do: [:i |
!     sidx := oclasses indexOf: (oclasses at: i) superClass.
!     pos := failedStream position.
!     nclasses add: (
!       (oclasses at: i)
!         copyClassWithSuperClass: (nclasses at: sidx)
!         errorsTo: failedStream).
!     dict := self dictionaryAndSymbolOf: (oclasses at: i).
!     autoLock ifTrue: [
!       assn := (dict at: 1) associationAt: (dict at: 2).
!       System writeLock: assn ifDenied: lblock ifChanged: lblock ifNotCommitted: [].
!     ].
!     dict == nil ifTrue: [
!       System myUserGlobals at: (nclasses at: i) name put: (nclasses at: i)
!     ]
!     ifFalse: [
!       (dict at: 1) at: (dict at: 2) put: (nclasses at: i)
!     ].
!     self linkOldClass: (oclasses at: i) toNewClass: (nclasses at: i).
!     pos ~= failedStream position ifTrue: [
!       badclasses add: (oclasses at: i)
!     ].
!   ].
! ].
! failedMethods := failedStream contents.
! 1 to: nclasses size do: [:i |
!   self addClass: (nclasses at: i).
! ].
! 
! "Show any error messages"
! badclasses size > 0 ifTrue: [
!   badclasses do: [:each |
!     each logModified: 'method dictionary (failed recompilation)'
!   ].
!   x := nil.
! 
!   "signal compilation errors"
! 
! ].
! 
! badLocks size > 0 ifTrue: [
!   "display a notifier"
! ].
! 
! ^{ oclasses . nclasses . badclasses size > 0 }
! %

category: 'Reporting'
method: ClassOrganizer
referencesTo: aSymbol

"Returns an Array of two sequenceable collections.  The first contains
 GsMethods that refer to the given symbol, and the second contains
 corresponding indexes into sourceStrings where the first reference takes
 place."

^self referencesTo: aSymbol in: classes
%

! fix 41606
category: 'Reporting'
method: ClassOrganizer
referencesTo: aSymbol in: aclassSet

"Returns an Array of two sequenceable collections.  The first contains
 GsMethods that reference the given symbol as a literal, 
 and the second contains corresponding indexes into sourceStrings where 
 the first such reference takes place."

| result resulti cset cls methsSet sym val |

methsSet := IdentitySet new .
result := { } .
resulti := { } .
sym := Symbol _existingWithAll: aSymbol .
sym ifNotNil:[ 
  val := (GsCurrentSession currentSession resolveSymbol: sym) 
              ifNotNil:[ :assoc| assoc _value] .
  cset := aclassSet sortAscending.
  1 to: cset size do: [ :i |
    cls := cset at: i.
    2 timesRepeat:[ | mDict |
      mDict := cls _fullMethodDictEnv0 .
      mDict valuesDo:[ :method | 
	(method _literalsIncludesSymbol: sym value: val ) ifTrue:[
	   (methsSet _addIfAbsent: method) ifTrue:[
	     result add: method.
	     resulti add: (method sourceString findString: sym  startingAt: 1)
	   ].
	 ].
      ].
      cls := cls class .
    ].
  ].
].
^{ result . resulti }
%

category: 'Accessing'
method: ClassOrganizer
rootClass

"Returns the root class for this organizer."

^rootClass
%

category: 'Updating'
method: ClassOrganizer
rootClass: aClass

"Sets the root class of the receiver's hierarchy.  Not generally a useful
 thing to do."

rootClass := aClass
%

! fix bug 11544
category: 'Reporting'
method: ClassOrganizer
searchForCategory: catname in: classSet

  "Returns a collection of GsMethods in the given category."
| result cset catSym envId |
result := { } .
envId := 0 .
catname _validateByteClass: CharacterCollection.
catSym := Symbol _existingWithAll: catname .
catSym == nil ifTrue:[ ^ result ].
cset := classSet sortAscending: 'name' .
1 to: cset size do: [ :j |
  { cset at: j . (cset at: j) class } do: [ :cls | | cat mdict |
    cls categorysDo:[ :cat :selectors |
      cat == catSym ifTrue:[
	mdict := cls methodDictForEnv: envId .
	selectors do: [ :selector | | method |
	  (method := mdict at: selector otherwise: nil) ifNil:[ |emsg|
            emsg := 'Missing method ' , selector quoted ,
			 ' in category ' , catSym , ' in class ' , cls name.
            [
	      self notify: emsg   "notify: only available with Seaside/Ruby"
            ] onSynchronous: MessageNotUnderstood do:[:ex |
              self error: emsg
            ].
	  ] ifNotNil: [
	    result add: method
	  ].
	].
      ].
    ].
  ].
].
^result
%

category: 'Reporting'
method: ClassOrganizer
sendersOf: aSelector

"Returns an Array of two Arrays.  The first subarray contains GsMethods
 that send the given selector.  The second subarray contains indexes where the
 first use of the selector occurs within the sourceString of the method."

^self sendersOf: aSelector in: classes
%

category: 'Reporting'
method: ClassOrganizer
sendersOfReport: aSelector

"Returns a String describing the methods that are senders of the specified
 selector."

^ self _asReportString: ((self sendersOf: aSelector) at:1 ) 
%

method: ClassOrganizer
packageSendersOfReport: aSelector
  "Returns a String describing the methods that are senders of the 
   specified selector, and for which the methods' category begins with '*' "

^ self _asReportString:(
  ((self sendersOf: aSelector) at:1 ) select:[ :meth | 
       ((meth inClass categoryOfSelector: meth selector ) at: 1) == $*  ])
%

method: ClassOrganizer
pkgsends: aSelector
  ^ self packageSendersOfReport: aSelector
%

category: 'Reporting'
method: ClassOrganizer
sends: aSelector
  "a variant of sendersOfReport: easier to type when using topaz ."
  ^ self sendersOfReport: aSelector
%

method: ClassOrganizer
sendsC: aSelector
  "senders report sorted by method categories"
  ^ self _asCategoriesReportString: ((self sendersOf: aSelector) at:1 ) 
%

category: 'Reporting'
method: ClassOrganizer
sendersOf: aSelector in: aclassSet

"Returns an Array of two Arrays.  The first subarray contains GsMethods
 that send the given selector.  The second subarray contains indexes where the
 first use of the selector occurs within the sourceString of the method."

| result resulti cset cls aSymbol methsSet |

methsSet := IdentitySet new .
result := { } .
resulti := { } .
aSymbol := Symbol _existingWithAll: aSelector .
aSymbol == nil ifTrue:[ ^{ result . resulti } ].

(aclassSet isKindOf: ClassSet) ifFalse: [
  cset := ClassSet withAll: aclassSet
]
ifTrue: [
  cset := aclassSet
].

cset := cset sortAscending.
1 to: cset size do: [ :i |
  cls := cset at: i.
  2 timesRepeat:[ | mDict |
    mDict := cls _fullMethodDictEnv0 .
    mDict valuesDo:[ :gsMethod | | srcOffset |
      srcOffset := gsMethod _sourceOffsetOfFirstSendOf: aSymbol .
      srcOffset ~~ nil ifTrue:[
        (methsSet _addIfAbsent: gsMethod) ifTrue:[
          result add: gsMethod .
          resulti add: srcOffset.
        ].
      ].
    ].
    cls := cls class .
  ].
].
^{ result . resulti }
%

method: ClassOrganizer
accessorsOf: ivName inClass: aClass
  | allSubs arr nam |
  nam := ivName asSymbol .
  allSubs := self allSubclassesOf: aClass .
  allSubs add: aClass .
  arr := { } .
  allSubs do:[:cls | 
    cls selectors do:[:sel | | meth |
      meth := cls compiledMethodAt: sel .
      (meth instVarsAccessed includes: nam) ifTrue:[ arr add: meth ].
    ].
  ].
  ^ arr 
%  

method: ClassOrganizer
accessorsReport: ivName inClass: aClass
  ^ self _asReportString:( self accessorsOf: ivName inClass: aClass)
%

category: 'Reporting'
method: ClassOrganizer
sentButNotImplemented
 "Returns an Array of selectors which are sent but not implemented"

  | sent implemented |
  sent := IdentitySet new.
  implemented := IdentitySet new .
  classes do:[ :aClass | | cls |
    cls := aClass .
    2 timesRepeat:[ | mDict |
      mDict := cls _fullMethodDictEnv0 .
      mDict valuesDo:[ :method | 
        implemented add:  method selector .
        sent addAll:  method _selectorPool
      ].
      cls := cls class .
    ].
  ].
  ^ (sent - implemented) sortAscending:''
%

method: ClassOrganizer
sentButNotImplementedReport
  "Returns a report of selectors sent but not implemented."
  | report LF |
  report := String new .
  LF := Character lf .
  self sentButNotImplemented do:[:sel | | methsRpt |
    report addAll: sel ; add: LF  .
    methsRpt := self _asReportString:((self sendersOf: sel) at:1) indent: '   '.
    methsRpt size > 0 
      ifTrue:[  report addAll: methsRpt ]
      ifFalse:[ report addAll: '   <no sender methods found>'; add: LF ].
  ].
  ^ report 
%
method: ClassOrganizer
sendsNotImpl

 ^ self sentButNotImplementedReport
%

category: 'Modifying Classes'
method: ClassOrganizer
makeInstancesPersistent: aClass 

"Recursively make aClass and all subclasses persistable.

 If an error occurs, the session will be unable to commit and
 must logout/login before commits are allowed again."

<primitive: 2001> 
| prot |
prot := System _protectedMode .
[
  aClass instancesNonPersistent ifTrue:[ | success |
    [ | subCs |
      success := false .
      aClass _makeInstancesPersistent .
      subCs := self subclassesOf: aClass .
      subCs do:[ :aSubcls |
	self makeInstancesPersistent: aSubcls .
      ].
      success := true .
    ] ensure:[
      success ifFalse:[ System _disallowCommitClassModFailure ].
    ]
  ].
] ensure:[
  prot _leaveProtectedMode
]
%

category: 'Modifying Classes'
method: ClassOrganizer
makeInstancesNonPersistent: aClass 

"Recursively make aClass and all subclasses non-persistent.

 If an error occurs, the session will be unable to commit and
 must logout/login before commits are allowed again."

<primitive: 2001> 
| prot |
prot := System _protectedMode .
[
  aClass instancesNonPersistent ifFalse:[ | success |
    [ | subCs |
      success := false .
      aClass _makeInstancesNonPersistent .
      subCs := self subclassesOf: aClass .
      subCs do:[ :aSubcls |
	self makeInstancesNonPersistent: aSubcls .
      ].
      success := true .
    ] ensure:[
      success ifFalse:[ System _disallowCommitClassModFailure ].
    ]
  ].
] ensure:[
  prot _leaveProtectedMode
]
%

category: 'Queries'
method: ClassOrganizer
subclassesOf: aClass

"Returns a copy of the set of subclasses for the given class.
 Generates an error if the receiver does not hold the given class."

classes == nil ifTrue: [
  self updateClassInfo
].
(classes includesIdentical: aClass) ifFalse: [
  self addClass: aClass.
  (classes includesIdentical: aClass) ifFalse: [
    self addClass: aClass.
    ^self subclassesOf: aClass
  ].
].
^(hierarchy at: aClass ifAbsent: [ClassSet new]) copy
%

category: 'Reporting'
method: ClassOrganizer
subclassesReport: aClass includeOops: withOopsBool
 | sorted str |
 sorted := SortedCollection sortBlock: [ :a :b | a name <= b name ].
 (self subclassesOf: aClass) do:[:aClass | sorted add: aClass ].
 str := String new .
 sorted do:[ :cls |  
   str add: cls name .
   withOopsBool ifTrue:[ str add: '   '; add: cls asOop asString ].
   str lf .
 ].
 ^ str
%

method: ClassOrganizer
subclassesReport: aClass
  ^ self subclassesReport: aClass includeOops: false
%

category: 'Reporting'
method: ClassOrganizer
substringSearch: aString

"Returns an Array of two Arrays.  The first subarray contains GsMethods
 whose sources include the given substring.  The second subarray contains
 indexes where the first occurrence of the substring was found."

^self substringSearch: aString in: classes
%

category: 'Reporting'
method: ClassOrganizer
substringReport: aString

"Returns a String describing the methods whose source contains
 the specified string ."

^ self _asReportString:
      ((self _substringSearch: aString in: classes ignoreCase: false) at: 1)
%

category: 'Reporting'
method: ClassOrganizer
strings: aString

"A variant of substringReport:  easier to type."

^ self _asReportString:
      ((self _substringSearch: aString in: classes ignoreCase: false) at: 1)
% 
method: ClassOrganizer
stringsC: aString

"A variant of substringReport:  sorted by method categories"

^ self _asCategoriesReportString:
      ((self _substringSearch: aString in: classes ignoreCase: false) at: 1)
% 

category: 'Reporting'
method: ClassOrganizer
stringsIc: aString

"A case-insensitive variant of substringReport:  easier to type."

^ self _asReportString:
      ((self _substringSearch: aString in: classes ignoreCase: true) at: 1)
% 
method: ClassOrganizer
stringsIcC: aString

"A case-insensitive variant of substringReport:  sorted by method categories"

^ self _asCategoriesReportString:
      ((self _substringSearch: aString in: classes ignoreCase: true) at: 1)
%

category: 'Reporting'
method: ClassOrganizer
substringSearch: aString in: aclassSet 

"Returns an Array of two Arrays.  The first subarray contains GsMethods
 whose sources include the given substring.  The second subarray contains
 indexes where the first occurrence of the substring was found."

^ self _substringSearch: aString in: aclassSet ignoreCase: false
%

category: 'Reporting'
method: ClassOrganizer
_substringSearch: aString in: aclassSet ignoreCase: icBoolean
"Returns an Array of two Arrays.  The first subarray contains GsMethods
 whose sources include the given substring.  The second subarray contains
 indexes where the first occurrence of the substring was found."

| result resulti list cls methsSet |
methsSet := IdentitySet new .
result := { } .
resulti := { } .
list := aclassSet sortAscending.
1 to: list size do: [ :i |
  cls := list at: i.
  2 timesRepeat:[ | mDict |
    mDict := cls _fullMethodDictEnv0 .
    mDict valuesDo:[ :gsMethod | | index |
        index := gsMethod sourceString _findString: aString startingAt: 1 
				      ignoreCase: icBoolean .
        index > 0 ifTrue: [
          (methsSet _addIfAbsent: gsMethod) ifTrue:[
            result add: gsMethod .
            resulti add: index
          ].
	].
    ].
    cls := cls class .
  ].
].
^{ result . resulti }
%

category: 'Class Collection'
method: ClassOrganizer
update

"Causes the receiver to rescan for classes and rebuild internal
 structures.  Synonymous with updateClassInfo."

self updateClassInfo
%

category: 'Class Collection'
method: ClassOrganizer
updateClassInfo

"Causes the receiver to rescan for classes and rebuild internal structures."

self collectClasses; "find all visible classes"
    rebuildHierarchy; "build the hierarchy, fleshing out class set too"
    collectClassNames; "pull out the names of the classes and form auto-complete set"
    rebuildCategories "build the class categories"
%

category: 'Reporting'
method: ClassOrganizer
hierarchyReport

"Returns a String that is a class hierarchy report for all classes
 known to the receiver."

| report |
report := String new .
self _hierarchyReportForClass: Object indent: '' report: report .
^ report
%

category: 'Reporting'
method: ClassOrganizer
referencesToObject: anObject

"Returns an Array of GsMethods that reference the given object
 (through a variable binding)"

^self referencesToObject: anObject in: classes
%

category: 'Reporting'
method: ClassOrganizer
referencesToObject: anObject in: aclassSet

"Returns an Array of GsMethods that reference the given object
 (through a variable binding)"

| result cset cls |

result := { } .

cset := aclassSet sortAscending.
1 to: cset size do: [ :i |
  cls := cset at: i.
  2 timesRepeat:[ | mDict |
    mDict := cls _fullMethodDictEnv0 .
    mDict valuesDo:[ :method | 
        (method _literalsIncludesValue: anObject) ifTrue:[
           result add: method 
        ].
    ].
    cls := cls class .
  ].
].
^result
%

category: 'Private'
method: ClassOrganizer
_hierarchyReportForClass: aClass indent: indent report: report 

""

| subClsArray nextIndent |
report addAll: indent; add: aClass name ; add: Character lf .
subClsArray := 
  (hierarchy at: aClass ifAbsent:[ IdentitySet new ]) sortAscending: 'name' .
nextIndent := indent , '  ' .
subClsArray do:[ :aSubCls |
  self _hierarchyReportForClass: aSubCls indent: nextIndent report: report 
  ].
%
method: ClassOrganizer
_hierarchyIvReportForClass: aClass indent: indent report: report
| subClsArray nextIndent ivNames |
report addAll: indent; add: aClass name .
ivNames := aClass _instVarNamesWithSeparator:  ''   .
ivNames size ~~ 0 ifTrue: [ report add: $(; add: ivNames ; add: $) ].
report lf .
subClsArray :=
  (hierarchy at: aClass ifAbsent:[ IdentitySet new ]) sortAscending: 'name' .
nextIndent := indent , '  ' .
subClsArray do:[ :aSubCls |
  self _hierarchyIvReportForClass: aSubCls indent: nextIndent report: report
  ].
%

method: ClassOrganizer
_hierarchyIvReport: aClass
| str |
str := String new .
self _hierarchyIvReportForClass: aClass indent: '' report: str .
^ str
%
